@@ -22,12 +22,12 @@
 
 ;;; Commenty:
 ;;
-;; The Emacs/w3 package (url) already provides WebDAV interface for Emacsen.
-;; But Emacs/w3 is too big for emacs-w3m users.
+;; The w3 package (url) already provides WebDAV interface for Emacsen.
+;; But w3 is too big for non-w3 users such as emacs-w3m users.
 ;;
 ;; This program currently uses
 ;; `nd - Tiny little command line WebDAV interface' as an WebDAV
-;; backend program. it is available at: http://www.gohome.org/nd/
+;; backend program. It is available at: http://www.gohome.org/nd/
 ;;
 ;;
 ;; Usage:
@@ -48,23 +48,37 @@
 ;; See http://www.gohome.org/eldav for more details.
 ;;
 
+(eval-when-compile
+  (require 'cl))
+
+;; For make-temp-file
+(eval-and-compile
+  (or (fboundp 'make-temp-file)
+      (require 'poe)))
+
+;; For non-mule XEmacs.
+(if (and (featurep 'xemacs)
+	 (not (featurep 'mule)))
+    (require 'poem))
+
+(require 'timezone)
+
+(defmacro eldav-static-if (cond then &rest else)
+  "Like `if', except that it evaluates COND at compile-time."
+  (if (eval cond) then (` (progn  (,@ else)))))
+(put 'eldav-static-if 'lisp-indent-function 2)
+
 (defgroup eldav nil
   "Accessing remote files and directories using WebDAV."
   :group 'files
   :prefix "eldav-")
 
 (defcustom eldav-debug nil
-  "*If non-nil, process string is inserted to the debug buffer."
+  "If non-nil, process string is inserted to the debug buffer."
   :type 'boolean
   :group 'eldav)
 
-(defcustom eldav-temporal-directory
-  (file-name-as-directory
-   (cond ((fboundp 'temp-directory)
-	  (temp-directory))
-	 ((boundp 'temporary-file-directory)
-	  temporary-file-directory)
-	 (t "/tmp")))
+(defcustom eldav-temporal-directory temporary-file-directory
   "Name of temporal directory."
   :group 'eldav
   :type 'directory)
@@ -79,8 +93,17 @@
   :group 'eldav
   :type 'regexp)
 
-(defcustom eldav-lock-identifier (concat "mailto:" user-mail-address)
-  "*URL used as contact information when creating locks in DAV.
+(defcustom eldav-lock-identifier (or (and (boundp 'user-mail-address)
+					  user-mail-address
+					  (> (length user-mail-address) 0)
+					  (concat "mailto:"
+						  user-mail-address))
+				     (and (getenv "USER")
+					  (concat "mailto:"
+						  (getenv "USER")
+						  "@" (system-name)))
+				     (system-name))
+  "URL used as contact information when creating locks in DAV.
 This will be used as the contents of the DAV:owner/DAV:href tag to
 identify the owner of a LOCK when requesting it.  This will be shown
 to other users when the DAV:lockdiscovery property is requested, so
@@ -106,7 +129,17 @@ Don't use any other value."
 (defcustom eldav-process-connection-type
   (not (and (featurep 'xemacs)
 	    (string-match "solaris" system-configuration)))
-  "*Process connection type for command execution."
+  "Process connection type for command execution."
+  :group 'eldav
+  :type 'boolean)
+
+(defcustom eldav-executable-property-namespace "http://apache.org/dav/props/"
+  "The namespace for `executable' property."
+  :group 'eldav
+  :type 'string)
+
+(defcustom eldav-use-vc nil
+  "VC is used if non-nil."
   :group 'eldav
   :type 'boolean)
 
@@ -121,6 +154,25 @@ Don't use any other value."
 (defvar eldav-last-propfind-result nil
   "A cache to save the last propfind result.")
 
+(defvar eldav-running-process nil)
+(make-variable-buffer-local 'eldav-running-process)
+(defvar eldav-process-auth-required nil)
+(make-variable-buffer-local 'eldav-process-auth-required)
+(defvar eldav-process-url nil)
+(make-variable-buffer-local 'eldav-process-url)
+(defvar eldav-process-user nil)
+(make-variable-buffer-local 'eldav-process-user)
+(defvar eldav-process-realm nil)
+(make-variable-buffer-local 'eldav-process-realm)
+(defvar eldav-process-passwd nil)
+(make-variable-buffer-local 'eldav-process-passwd)
+(defvar eldav-process-proxy-user nil)
+(make-variable-buffer-local 'eldav-process-proxy-user)
+(defvar eldav-process-proxy-passwd nil)
+(make-variable-buffer-local 'eldav-process-proxy-passwd)
+(defvar eldav-process-proxy-realm nil)
+(make-variable-buffer-local 'eldav-process-proxy-realm)
+
 (defvar eldav-debug-buffer nil)
 (defun eldav-debug (string)
   "Insert STRING to the debug buffer."
@@ -186,7 +238,8 @@ Don't use any other value."
 (defun eldav-file-inode-number (name)
   (let ((inode (eldav-get-hash-entry name eldav-inode-hashtable)))
     (or inode
-	(progn (eldav-put-hash-entry name (incf eldav-inode-max) eldav-inode-hashtable)
+	(progn (eldav-put-hash-entry name (incf eldav-inode-max)
+				     eldav-inode-hashtable)
 	       eldav-inode-max))))
 
 (defun eldav-url-server-root (url)
@@ -202,19 +255,22 @@ Don't use any other value."
 		(length (eldav-file-name-server-path name)))))
 
 (defun eldav-file-name-server-path (name)
-  (let ((url (eldav-file-name-url name)))
-    (if (string-match (concat "^" (regexp-quote 
-				   (concat "http://"
-					   (eldav-url-server-root url))))
-		      url)
-	(eldav-url-decode-string
-	 (substring url (match-end 0))))))
+  (let ((url (eldav-file-name-url name))
+	server-path)
+    (when (string-match (concat "^" (regexp-quote
+				     (concat "http://"
+					     (eldav-url-server-root url))))
+			url)
+      (setq server-path (eldav-url-decode-string
+			 (substring url (match-end 0))))
+      (if (zerop (length server-path)) "/" server-path))))
 
-(defun eldav-lockinfo (url)
-  "Lock information of URL."
-  (if (string-match "^.*/$" url)
-      (setq url (substring url 0 (- (length url) 1))))
-  (cdr (assq 'lock (cdr (car (eldav-propfind url))))))
+(defun eldav-lockinfo-list (url)
+  "Lock information list of URL."
+  (let ((eldav-inhibit-propfind-cache t))
+    (if (string-match "^.*/$" url)
+	(setq url (substring url 0 (- (length url) 1))))
+    (cdr (assq 'lock-list (cdr (car (ignore-errors (eldav-propfind url))))))))
 
 (defun eldav-lockinfo-token (lockinfo)
   "Return token of LOCKINFO."
@@ -238,11 +294,24 @@ Don't use any other value."
   "Return scope of LOCKINFO."
   (cadr (assq 'scope lockinfo)))
 
+(defun eldav-lockinfo-list-search (name value list)
+  "Return the matched lockinfo list which matches its NAME and VALUE.
+LIST is the lockinfo list."
+  (let (match)
+    (while list
+      (when (equal (apply (intern (concat "eldav-lockinfo-"
+					  (symbol-name name)))
+			  (list (car list)))
+		   value)
+	(setq match (cons (car list) match)))
+      (setq list (cdr list)))
+    match))
+
 ;;; file-name prefix definition.
 (defconst eldav-file-name-prefix ":http://")
 (defconst eldav-file-name-regexp (concat "^"
 					 (regexp-quote eldav-file-name-prefix)
-					 "[^/]+/"))
+					 "\\([^/]+\\)"))
 (add-to-list 'file-name-handler-alist
 	     (cons eldav-file-name-regexp 'eldav-handler))
 
@@ -288,8 +357,8 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	(enable-recursive-minibuffers t)
 	passwd)
     (or (cdr (assoc username
-		    (cdr (assoc realm 
-				(assoc root 
+		    (cdr (assoc realm
+				(assoc root
 				       eldav-process-authinfo-alist)))))
 	(progn
 	  (eldav-process-set-authinfo
@@ -304,20 +373,23 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	 (ident (or realm root))
 	 (enable-recursive-minibuffers t)
 	 (alist))
-    (if (setq alist
-	      (cdr (assoc realm
-			  (cdr (assoc root
-				      eldav-process-authinfo-alist)))))
-	(if (= 1 (length alist))
-	    (caar alist)
-	  (completing-read (if ident
-			       (format "Select username for %s: " ident)
-			     "Select username: ")
-			   (mapcar (lambda (x) (cons (car x) (car x))) alist)
-			   nil t))
-      (read-from-minibuffer (if ident
-				(format "Username for %s: " ident)
-			      "Username: ")))))
+    (condition-case nil
+	(if (setq alist
+		  (cdr (assoc realm
+			      (cdr (assoc root
+					  eldav-process-authinfo-alist)))))
+	    (if (= 1 (length alist))
+		(caar alist)
+	      (completing-read (if ident
+				   (format "Select username for %s: " ident)
+				 "Select username: ")
+			       (mapcar (lambda (x) (cons (car x) (car x)))
+				       alist)
+			       nil t))
+	  (read-from-minibuffer (if ident
+				    (format "Username for %s: " ident)
+				  "Username: ")))
+      (quit (delete-process eldav-running-process)))))
 
 (defun eldav-process-clear-authinfo (url realm username)
   (eldav-process-set-authinfo url realm username nil))
@@ -336,13 +408,22 @@ If PASSWORD is nil, remove authinfo entry if exists."
       (when (string-match (concat "^" (regexp-quote (car elem))) url)
 	(throw 'found (cdr elem))))))
 
+(eval-when-compile
+  (defvar system-time-locale)
+  (defvar message-log-max)
+  (defvar coding-system-used)
+  (defvar default-process-coding-system))
+
+(defun eldav-process-sentinel (process event)
+  ;; nothing to do.
+  )
+
 (defun eldav-process-filter (process string)
   (eldav-debug string)
   (with-current-buffer (process-buffer process)
     (let ((buffer-read-only nil)
 	  (case-fold-search nil)
-	  (enable-recursive-minibuffers t)
-	  clear realm passwd)
+	  (enable-recursive-minibuffers t))
       (goto-char (process-mark process))
       (insert string)
       (unless eldav-process-auth-required
@@ -369,9 +450,9 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	    (setq eldav-process-proxy-user
 		  (read-from-minibuffer (match-string 0))))
 	  (ignore-errors
+	    (erase-buffer)
 	    (process-send-string process (concat eldav-process-proxy-user
-						 "\n"))
-	    (delete-region (point-min) (point-max))))
+						 "\n"))))
 	 ((and eldav-process-auth-required
 	       (looking-at "Proxy Password: ")
 	       (= (match-end 0) (point-max)))
@@ -379,9 +460,9 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	    (setq eldav-process-proxy-passwd
 		  (read-passwd (match-string 0))))
 	  (ignore-errors
+	    (erase-buffer)
 	    (process-send-string process (concat eldav-process-proxy-passwd
-						 "\n"))
-	    (delete-region (point-min) (point-max))))
+						 "\n"))))
 	 ((and eldav-process-auth-required
 	       (looking-at "Password: ")
 	       (= (match-end 0) (point-max)))
@@ -390,8 +471,8 @@ If PASSWORD is nil, remove authinfo entry if exists."
 					   eldav-process-realm
 					   eldav-process-user))
 	  (ignore-errors
-	    (process-send-string process (concat eldav-process-passwd "\n"))
-	    (delete-region (point-min) (point-max))))
+	    (erase-buffer)
+	    (process-send-string process (concat eldav-process-passwd "\n"))))
 	 ((and eldav-process-auth-required
 	       (looking-at "\n?Username for \\(.*\\): ")
 	       (= (match-end 0) (point-max)))
@@ -399,7 +480,7 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	  (when (eq (string-to-char eldav-process-realm) ?\")
 	    (setq eldav-process-realm
 		  (read eldav-process-realm)))
-	  (when eldav-process-user	; Second time.
+	  (when eldav-process-user ; Second time, which means wrong auth.
 	    (eldav-process-clear-authinfo eldav-process-url
 					  eldav-process-realm
 					  eldav-process-user))
@@ -407,34 +488,9 @@ If PASSWORD is nil, remove authinfo entry if exists."
 		(eldav-process-read-user eldav-process-url
 					 eldav-process-realm))
 	  (ignore-errors
-	    (process-send-string process (concat eldav-process-user "\n"))
-	    (delete-region (point-min) (point-max)))))))))
-				   
-(defun eldav-process-sentinel (process event)
-  (with-current-buffer (process-buffer process)
-    (setq eldav-process-running nil)
-    (if eldav-running-process
-	(delete-process eldav-running-process))))
-
-(defvar eldav-running-process nil)
-(make-variable-buffer-local 'eldav-running-process)
-
-(defvar eldav-process-auth-required nil)
-(defvar eldav-process-url nil)
-(make-variable-buffer-local 'eldav-process-url)
-(defvar eldav-process-user nil)
-(make-variable-buffer-local 'eldav-process-user)
-(defvar eldav-process-realm nil)
-(make-variable-buffer-local 'eldav-process-realm)
-(defvar eldav-process-passwd nil)
-(make-variable-buffer-local 'eldav-process-passwd)
-(defvar eldav-process-proxy-user nil)
-(make-variable-buffer-local 'eldav-process-proxy-user)
-(defvar eldav-process-proxy-passwd nil)
-(make-variable-buffer-local 'eldav-process-proxy-passwd)
-(defvar eldav-process-proxy-realm nil)
-(make-variable-buffer-local 'eldav-process-proxy-realm)
-
+	    (erase-buffer)
+	    (process-send-string process (concat eldav-process-user
+						 "\n")))))))))
 
 (defun eldav-exec (url &rest args)
   "Exec nd command for URL with ARGS."
@@ -444,9 +500,10 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	(default-process-coding-system (cons 'binary 'binary))
 	(process-connection-type eldav-process-connection-type)
 	(enable-recursive-minibuffers t)
-	(realm (eldav-url-realm url)))
+	(realm (eldav-url-realm url))
+	proc)
     (with-current-buffer (get-buffer-create " *eldav*")
-      (unless (featurep 'xemacs) (set-buffer-multibyte nil))
+      (eldav-static-if (featurep 'xemacs) nil (set-buffer-multibyte nil))
       (if (and eldav-proxy
 	       (or (null eldav-no-proxy-url-regexp)
 		   (not (string-match eldav-no-proxy-url-regexp url))))
@@ -461,16 +518,24 @@ If PASSWORD is nil, remove authinfo entry if exists."
       (if eldav-process-proxy-realm
 	  (setq args (append args (list "-A"
 					eldav-process-proxy-realm))))
+      ;;      (eldav-debug (concat "EXEC: "
+      ;;			   eldav-program-name
+      ;;			   " "
+      ;;			   (mapconcat
+      ;;			    'identity
+      ;;			    (append eldav-program-arguments args
+      ;;				    (list url))
+      ;;			    " ")))
       (setq proc (apply 'start-process
 			"eldav-exec"
 			(current-buffer)
-			eldav-program-name 
+			eldav-program-name
 			(append eldav-program-arguments args
 				(list url))))
       (setq eldav-running-process proc
 	    eldav-process-user nil
 	    eldav-process-realm nil
-	    eldav-process-password nil
+	    eldav-process-passwd nil
 	    eldav-process-url url)
       (set-process-filter proc 'eldav-process-filter)
       (set-process-sentinel proc 'eldav-process-sentinel)
@@ -485,18 +550,18 @@ If PASSWORD is nil, remove authinfo entry if exists."
 	       (setq proc nil)))
       (when proc
 	(accept-process-output proc 0 200))
-      (setq eldav-running-process nil)
       (message "")
+      (setq eldav-running-process nil)
       (if proc
 	  (if (and (numberp (process-exit-status proc))
 		   (zerop (process-exit-status proc)))
 	      (buffer-string)
 	    (if (numberp (process-exit-status proc))
 		(let ((err (ignore-errors (read (buffer-string)))))
-		  (if (eq (car err) 'error)
+		  (if (and (consp err)
+			   (eq (car err) 'error))
 		      (eval err)
-		    (error "Status %d on DAV process"
-			   (process-exit-status proc))))
+		    (error "DAV process is stopped")))
 	      (error "Error on DAV process `%s'"
 		     (process-exit-status proc))))
 	(error "DAV process was stopped.")))))
@@ -511,7 +576,7 @@ If PASSWORD is nil, remove authinfo entry if exists."
     (let ((props (read (condition-case err
 			   (apply 'eldav-exec url (list "-v"))
 			 (error
-			  (setq eldav-last-propfind-result (cons url nil))
+			  (setq eldav-last-propfind-result nil)
 			  (signal (car err) (cdr err)))))))
       (setq props (mapcar (lambda (prop)
 			    (setcar prop
@@ -521,29 +586,36 @@ If PASSWORD is nil, remove authinfo entry if exists."
       (setq eldav-last-propfind-result (cons url props))
       props)))
 
+(defun eldav-proppatch (url prop value &optional namespace)
+  "Apply PROPPATCH to the URL.
+PROP is the name of the property.
+VALUE is the value for the property.
+Optional NAMESPACE is the namespace for the property."
+  (apply 'eldav-exec url
+	 (append (list "-e" (concat prop "=" value))
+		 (if namespace (list "-N" namespace)))))
+
 (defun eldav-get (url)
   "Get URL."
   (eldav-exec url))
 
 (defun eldav-mkcol (url &optional token)
   "MKCOL URL."
-  (apply 'eldav-exec url 
+  (apply 'eldav-exec url
 	 (append (list "-k")
 		 (if token (list "-t" token)))))
 
-;; No use?
 (defun eldav-lock (url &optional owner scope timeout)
   "Lock URL with OWNER, SCOPE and TIMEOUT."
   (read (apply 'eldav-exec url
 	       (append (list "-l")
-;; nd does not accept -o owner
-;;		       (list "-o" (or owner 
-;;				      eldav-lock-identifier))
+		       (list "-o" (or owner
+				      eldav-lock-identifier))
 		       (if scope (list "-s" (symbol-name scope)))
 		       (if (and timeout
 				(not (eq timeout 'infinite))
 				(numberp timeout))
-			   (list 
+			   (list
 			    "-i"
 			    (format "Second-%d" timeout)))))))
 
@@ -558,21 +630,24 @@ If PASSWORD is nil, remove authinfo entry if exists."
 If BUFFER is specified its content is written.
 Default is current buffer.
 Use lock TOKEN if specified."
-  (let ((temp-file (eldav-real-expand-file-name
-		    (make-temp-name "eldav-el")
-		    eldav-temporal-directory))
+  (let ((temp-file (make-temp-file
+		    (eldav-real-expand-file-name
+		     "eldav-el"
+		     eldav-temporal-directory)))
 	(coding-system-for-write 'binary)
 	status)
+    ;; for Emacs 21.1-21.3
+    (set-file-modes temp-file 384)
     (let (default-directory)
       (eldav-real-write-region (point-min) (point-max) temp-file nil 'no-msg)
       (unwind-protect
-	  (setq status 
-		(read (apply 'eldav-exec url (append 
+	  (setq status
+		(read (apply 'eldav-exec url (append
 					      (list "-p" temp-file)
 					      (if token (list "-t" token))))))
 	(delete-file temp-file))
       (eldav-check-multistat "PUT" url status))))
-      
+
 (defun eldav-check-multistat (method url status)
   (dolist (stat status)
     (let (code)
@@ -581,11 +656,11 @@ Use lock TOKEN if specified."
 	(setq code (cadr (assq 'status (cdr stat))))
 	(if (> code 300)
 	    (cond
-	     ((eq code 422) 
+	     ((eq code 422)
 	      (error "%s failed, `Unprocessable Entity'" method))
-	     ((eq code 423) 
+	     ((eq code 423)
 	      (error "%s failed, `Locked'" method))
-	     ((eq code 424) 
+	     ((eq code 424)
 	      (error "%s failed, `Failed Dependency'" method))
 	     (t (error "%s failed, code=`%d'" method code))))))))
 
@@ -624,11 +699,12 @@ Use lock TOKEN if specified."
 
 (defun eldav-copy-file (file newname &optional ok-if-already-exists keep-date)
   (interactive "fCopy file: \nFCopy %s to file: \np")
-  (if (eq (find-file-name-handler newname 'copy-file) 'eldav-handler)
+  (if (and (eq (find-file-name-handler newname 'copy-file) 'eldav-handler)
+	   (eq (find-file-name-handler file 'copy-file) 'eldav-handler))
       (eldav-copy (eldav-file-name-url file)
 		  (eldav-file-name-url newname) ok-if-already-exists)
     (with-temp-buffer
-      (unless (featurep 'xemacs) (set-buffer-multibyte nil))
+      (eldav-static-if (featurep 'xemacs) nil (set-buffer-multibyte nil))
       (let ((coding-system-for-read 'binary))
 	(insert-file-contents file))
       (if (and (not ok-if-already-exists) (file-exists-p newname))
@@ -642,19 +718,18 @@ Use lock TOKEN if specified."
 
 (defun eldav-delete-file (file)
   (unless (eldav-file-directory-p file)
-    (if (file-exists-p file)
-	(eldav-delete (eldav-file-name-url file))
-      (signal 'file-error (list "Removing old name" file)))))
+    (let ((eldav-inhibit-propfind-cache t))
+      (if (file-exists-p file)
+	  (eldav-delete (eldav-file-name-url file))
+	(signal 'file-error (list "Removing old name" file))))))
 
 (defun eldav-directory-files (dir &optional full match nosort)
   (let ((server-path (eldav-file-name-server-path dir))
-	node-file isdir files)
+	node-file files)
     (dolist (node (eldav-propfind (eldav-file-name-url dir)))
-      (setq node-file (car node)
-	    isdir nil)
+      (setq node-file (car node))
       (unless (string= server-path node-file)
 	(when (string-match "\\(.*\\)/$" node-file)
-	  (setq isdir t)
 	  (setq node-file (match-string 1 node-file)))
 	(setq node-file (file-name-nondirectory node-file))
 	(if (and match (not (string-match match node-file)))
@@ -665,7 +740,7 @@ Use lock TOKEN if specified."
 			      (expand-file-name file dir)) files)))
     (if nosort
 	files
-      (sort files 'string-lessp))))
+      (sort (copy-sequence files) 'string-lessp))))
 
 (defun eldav-file-accessible-directory-p (dir)
   (eldav-file-directory-p dir))
@@ -678,12 +753,29 @@ Use lock TOKEN if specified."
 		   (aref x 2) (aref x 1) (aref x 0)
 		   (aref x 6)))))
 
+(defun eldav-node-property-value (node name &optional namespace)
+  "Get optional property value from NODE.
+NAME is the property name string.
+If optional NAMESPACE is specified, treat it as a namespace for the PROP."
+  (catch 'value
+    (dolist (prop (cdr node))
+      (when (and (eq (car prop) 'property)
+		 (string= (cadr (assq 'name prop)) name)
+		 (or (null namespace) (string= (cadr (assq 'ns prop))
+					       namespace)))
+	(throw 'value (cadr (assq 'value prop)))))))
+
 (defun eldav-node-attributes (node)
-  (let (last-modified isdir)
-    (when node
+  (let (last-modified isdir executable)
+    (when (and node
+	       (> (length (delq (assq 'lock-list (cdr node)) (cdr node))) 0))
       (if (or (string-match ".*/$" (car node))
 	      (string= (car node) ".."))
 	  (setq isdir t))
+      (setq executable (if (string=
+			    (eldav-node-property-value node "executable")
+			    "T")
+			   "x" "-"))
       (list isdir			;0 file type
 	    1				;1 link count
 	    -1				;2 uid
@@ -694,7 +786,8 @@ Use lock TOKEN if specified."
 		   (cadr (assq 'last-modified (cdr node))))) ;5 mtime
 	    last-modified		;6 ctime
 	    (cadr (assq 'size (cdr node))) ; 7 size
-	    (if isdir "drwxrwxrwx" "-rw-rw-rw-") ;8 mode
+	    (if isdir "drwxrwxrwx"
+	      (format "-rw%srw-rw-" executable)) ;8 mode
 	    nil				;9 gid weird
 	    (eldav-file-inode-number (car node)) ;10 "inode number".
 	    -1				;11 device number.
@@ -704,18 +797,28 @@ Use lock TOKEN if specified."
   (if (string-match eldav-file-name-regexp file)
       (progn
 	(setq file (expand-file-name file))
-	(eldav-node-attributes 
+	(eldav-node-attributes
 	 (car (ignore-errors (eldav-propfind (eldav-file-name-url file))))))
     (eldav-real-file-attributes file)))
 
+(defun eldav-file-props-directory-p (name props)
+  (string= (cadr (assq 'resourcetype (cdr (assoc
+					   (eldav-file-name-server-path
+					    (file-name-as-directory name))
+					   props)))) "collection"))
+
 (defun eldav-file-directory-p (name)
-  (let ((props (ignore-errors (eldav-propfind (eldav-file-name-url
-					       (file-name-as-directory 
-						name))))))
-    (and props t)))
+  (eldav-file-props-directory-p
+   name
+   (ignore-errors (eldav-propfind
+		   (eldav-file-name-url name)))))
 
 (defun eldav-file-executable-p (file)
-  nil)
+  (string=
+   (eldav-node-property-value
+    (car (ignore-errors (eldav-propfind (eldav-file-name-url file))))
+    "executable")
+   "T"))
 
 (defun eldav-file-exists-p (name)
   (condition-case nil
@@ -724,16 +827,21 @@ Use lock TOKEN if specified."
     (error)))
 
 (defun eldav-file-local-copy (file)
-  (let ((tempfile (expand-file-name (make-temp-name "eldav-el")
-				    eldav-temporal-directory)))
-    (with-temp-file tempfile
+  (let ((temp-file (make-temp-file
+		    (expand-file-name "eldav-el"
+				      eldav-temporal-directory))))
+    ;; for Emacs 21.1-21.3
+    (set-file-modes temp-file 384)
+    (with-temp-file temp-file
       (insert-file-contents file))
-    tempfile))
+    temp-file))
 
 (defun eldav-file-modes (file)
   (if (eldav-file-directory-p file)
       511
-    438))
+    (if (eldav-file-executable-p file)
+	502
+      438)))
 
 (defun eldav-file-name-all-completions (file dir)
   (let ((server-path (eldav-file-name-server-path dir))
@@ -753,8 +861,7 @@ Use lock TOKEN if specified."
       matches)))
 
 (defun eldav-file-name-completion (file dir)
-  (let ((server-path (eldav-file-name-server-path dir))
-	match isdir)
+  (let (isdir)
     (try-completion
      file
      (mapcar (lambda (node)
@@ -769,6 +876,10 @@ Use lock TOKEN if specified."
 
 (defsubst eldav-time-less-p (t1 t2)
   "Say whether time T1 is less than time T2."
+  (unless (consp (cdr t1))
+    (setq t1 (list (car t1)(cdr t1))))
+  (unless (consp (cdr t2))
+    (setq t2 (list (car t2)(cdr t2))))
   (or (< (car t1) (car t2))
       (and (= (car t1) (car t2))
 	   (< (nth 1 t1) (nth 1 t2)))))
@@ -785,8 +896,16 @@ Use lock TOKEN if specified."
 
 ;; `file-truename' is identity.
 
+(defun eldav-own-lock-p (file)
+  (let ((locks (eldav-lockinfo-list (eldav-file-name-url file))))
+    (member eldav-lock-identifier
+	    (mapcar 'eldav-lockinfo-owner locks))))
+
 (defun eldav-file-writable-p (file)
-  (file-exists-p (file-name-directory (directory-file-name file))))
+  ;; Read-only unless my lock is contained.
+  (if eldav-use-vc
+      (eldav-own-lock-p file)
+    (file-exists-p (file-name-directory (directory-file-name file)))))
 
 (defun eldav-insert-file-attribute-as-ls-line (name attr switches)
   (insert (nth 8 attr) " " (format "%3d" (or (nth 1 attr) 0))
@@ -806,23 +925,38 @@ Use lock TOKEN if specified."
 (defun eldav-insert-directory (file switches &optional wildcard full)
   (unless (string-match eldav-file-name-regexp file)
     (setq file (expand-file-name file default-directory)))
+  ;; remove '/' in the tail of the filename.
+  (when (string-match "\\(.*\\)/$" file)
+    (setq file (match-string 1 file)))
   (let ((server-path
 	 (file-name-as-directory (eldav-file-name-server-path file)))
-	(as-file (memq ?d (mapcar 'identity switches)))
-	node-file parent)
+	(as-file (catch 'done
+		   (dolist (option (split-string switches))
+		     (unless (string= option "--dired")
+		       (when (string-match "d" option)
+			 (throw 'done t))))))
+	node-file parent attr)
     (let ((parentdir (file-name-directory (directory-file-name file))))
       (setq parent (ignore-errors
-		     (eldav-propfind (eldav-file-name-url parentdir))))
-      (if parent (setcar parent "..")))
+		     (eldav-propfind
+		      (eldav-file-name-url parentdir))))
+      (when parent
+	(setq parent (copy-sequence parent))
+	(setcar parent "..")))
+    (when (and (file-directory-p file)
+	       (ignore-errors (eldav-propfind
+			       (eldav-file-name-url
+				(file-name-as-directory file)))))
+      (setq file (file-name-as-directory file)))
     (dolist (node
-	     (sort (append (if (and (not as-file) parent) (list parent))
-			   (eldav-propfind
-			    (eldav-file-name-url
-			     (if as-file
-				 (setq file (directory-file-name file))
-			       (if (file-directory-p file)
-				   (file-name-as-directory file)
-				 file)))))
+	     (sort (copy-sequence
+		    (append (if (and (not as-file) parent) (list parent))
+			    (let ((eldav-inhibit-propfind-cache t))
+			      (eldav-propfind
+			       (eldav-file-name-url
+				(if as-file
+				    (setq file (directory-file-name file))
+				  file))))))
 		   (lambda (n1 n2)
 		     (string-lessp
 		      (if (string= server-path (car n1))
@@ -832,7 +966,7 @@ Use lock TOKEN if specified."
 			  "."
 			(car n2))))))
       (setq node-file (car node))
-      (cond 
+      (cond
        ((string= node-file ".."))
        (t
 	(if (and (not as-file) (string= server-path node-file))
@@ -840,9 +974,10 @@ Use lock TOKEN if specified."
 	(when (string-match "\\(.*\\)/$" node-file)
 	  (setq node-file (match-string 1 node-file)))
 	(setq node-file (file-name-nondirectory node-file))))
-      (eldav-insert-file-attribute-as-ls-line node-file
-					      (eldav-node-attributes node)
-					      switches))))
+      (when (setq attr (eldav-node-attributes node))
+	(eldav-insert-file-attribute-as-ls-line node-file
+						attr
+						switches)))))
 
 (defun eldav-insert-file-contents (filename &optional visit beg end replace)
   (barf-if-buffer-read-only)
@@ -856,25 +991,35 @@ Use lock TOKEN if specified."
     (if beg (setq content (substring content beg)))
     (if end (setq content (substring content 0 (- end (or beg 0)))))
     (save-excursion
-      (insert 
-       (decode-coding-string 
+      (insert
+       (decode-coding-string
 	content
-	(if coding-system-for-read
-	    coding-system-for-read
-	  (or (catch 'done
-		(dolist (elem file-coding-system-alist)
-		  (when (string-match (car elem) filename)
-		    (setq coding (cdr elem))
-		    (if (and (symbolp (cdr elem))
-			     (fboundp (cdr elem)))
-			(setq coding (funcall coding)))
-		    (throw 'done (cond ((coding-system-p coding)
-					coding)
-				       ((consp coding)
-					(car coding)))))))
-	      'binary)))))
-    (when (boundp 'last-coding-system-used)
-      (setq coding-system-used last-coding-system-used))    
+	(cond
+	 ((not (boundp 'coding-system-for-read))
+	  nil)
+	 (coding-system-for-read
+	  coding-system-for-read)
+	 (t
+	  (eldav-static-if (fboundp 'find-operation-coding-system)
+	      (setq coding (find-operation-coding-system
+			    'insert-file-contents
+			    filename visit beg end replace))
+	    (catch 'done
+	      (dolist (elem file-coding-system-alist)
+		(when (string-match (car elem) filename)
+		  (setq coding (cdr elem))
+		  (if (and (symbolp (cdr elem))
+			   (fboundp (cdr elem)))
+		      (setq coding (funcall coding)))
+		  (throw 'done nil)))))
+	  (cond
+	   ((coding-system-p coding)
+	    coding)
+	   ((consp coding)
+	    (car coding))
+	   (t 'binary)))))))
+    (eldav-static-if (boundp 'last-coding-system-used)
+	(setq coding-system-used last-coding-system-used))
     (list filename (length content))))
 
 (defun eldav-load (file &optional noerror nomessage nosuffix)
@@ -913,6 +1058,15 @@ Use lock TOKEN if specified."
 (defun eldav-unhandled-file-name-directory (filename)
   (file-name-directory eldav-temporal-directory))
 
+(defun eldav-file-name-directory (filename)
+  (let ((name (eldav-real-file-name-directory filename)))
+    (if (string= eldav-file-name-prefix name)
+	(concat filename "/")
+      name)))
+
+(defun eldav-real-file-name-directory (&rest args)
+  (eldav-run-real-handler 'file-name-directory args))
+
 (defun eldav-real-write-region (&rest args)
   (eldav-run-real-handler 'write-region args))
 
@@ -931,9 +1085,11 @@ Use lock TOKEN if specified."
    ((string-match "/\\.$" name)
     (setq name (file-name-directory name))))
   (if (string-match eldav-file-name-regexp name)
-      (concat 
+      (concat
        (substring name 0 (match-end 0))
-       (substring (expand-file-name (substring name (match-end 0)) "/") 1))
+       (substring (expand-file-name (concat
+				     "/"
+				     (substring name (match-end 0)))) 1))
     ;; base name matches eldav path.
     (cond
      ((string= name ".")
@@ -963,29 +1119,40 @@ Use lock TOKEN if specified."
     (let ((file-mdtm (eldav-file-modtime name))
 	  (buf-mdtm (with-current-buffer buf (visited-file-modtime))))
       (or (zerop (car file-mdtm))
-	  (<= (float-time file-mdtm) (float-time buf-mdtm))))))
+	  (not (eldav-time-less-p buf-mdtm file-mdtm))))))
 
-(if (fboundp 'select-safe-coding-system)
+(eldav-static-if (fboundp 'select-safe-coding-system)
     (defun eldav-select-safe-coding-system (from to default-coding-system)
-      (select-safe-coding-system from to default-coding-system))
+      (funcall (symbol-function 'select-safe-coding-system)
+	       from to default-coding-system))
   (defun eldav-select-safe-coding-system  (from to default-coding-system)
-    default-coding-system))
+    (if (boundp 'default-coding-system)
+	default-coding-system)))
 
 (defun eldav-write-region (start end filename &optional append visit
 				 &rest args)
   (if append (signal 'file-error (list "Append write not supported")))
-  (let ((origin (current-buffer))
-	(coding buffer-file-coding-system))
+  (let* ((origin (current-buffer))
+	 (coding (if (boundp 'buffer-file-coding-system)
+		     buffer-file-coding-system))
+	 (url (eldav-file-name-url filename))
+	 (lockinfo (car (eldav-lockinfo-list-search
+			 'owner
+			 eldav-lock-identifier
+			 (eldav-lockinfo-list url)))))
     (setq filename (expand-file-name filename))
     (save-restriction
       (narrow-to-region start end)
       (with-temp-buffer
 	(insert-buffer origin)
 	(encode-coding-region (point-min) (point-max)
-			      (eldav-select-safe-coding-system 
+			      (eldav-select-safe-coding-system
 			       (point-min)
 			       (point-max) coding))
-	(eldav-put-buffer (eldav-file-name-url filename)))
+	(eldav-put-buffer url
+			  (current-buffer)
+			  (if lockinfo
+			      (eldav-lockinfo-token lockinfo))))
       (cond
        ((stringp visit))		; second file name?
        ((eq visit t)
@@ -996,7 +1163,9 @@ Use lock TOKEN if specified."
 	(message "Wrote %s" filename))
        ((null visit)
 	(message "Wrote %s" filename))))
-    (setq last-coding-system-used coding)))
+    (eldav-static-if (boundp 'last-coding-system-used)
+	(setq last-coding-system-used coding))
+    nil))
 
 ;;;###autoload
 (defun eldav-handler (operation &rest args)
@@ -1013,9 +1182,45 @@ Use lock TOKEN if specified."
 	(inhibit-file-name-operation operation))
     (apply operation args)))
 
-(or (memq 'eldav-set-buffer-mode find-file-hooks)
-    (setq find-file-hooks
-	  (cons 'eldav-set-buffer-mode find-file-hooks)))
+(defun eldav-vc-registered (file)
+  (when eldav-use-vc
+    (when (string-match eldav-file-name-regexp file)
+      (funcall (symbol-function 'vc-file-setprop) file 'vc-backend 'eldav)
+      t)))
+
+;;; Utils
+(defun eldav-set-file-executable-p (file value)
+  (let ((props (cdar (read (eldav-proppatch
+			    (eldav-file-name-url file)
+			    "executable" (if value "T" "F")
+			    eldav-executable-property-namespace)))))
+    (if (>= (cadr (assq 'status props)) 300)
+	(error "PROPPATCH Failed"))))
+
+;;;
+;; An example setting for dired:
+;; (add-hook 'dired-mode-hook
+;;	  (lambda ()
+;;	    (define-key
+;;	      dired-mode-map "\C-ce" 'dired-eldav-toggle-file-executable)))
+;;;###autoload
+(defun dired-eldav-toggle-file-executable (&optional arg)
+  "Set the executable property of marked (or next ARG) files."
+  (interactive "P")
+  (let* ((files (dired-get-marked-files t arg)))
+    (dolist (file files)
+      (setq file (expand-file-name file (dired-current-directory)))
+      (when (string-match eldav-file-name-regexp file)
+	(eldav-set-file-executable-p file (not (file-executable-p file)))))
+    (dired-do-redisplay arg)))
+
+(eldav-static-if (boundp 'find-file-hook)
+    (or (memq 'eldav-set-buffer-mode find-file-hook)
+	(add-hook 'find-file-hook
+		  'eldav-set-buffer-mode))
+  (or (memq 'eldav-set-buffer-mode find-file-hooks)
+      (setq find-file-hooks
+	    (cons 'eldav-set-buffer-mode find-file-hooks))))
 
 (put 'add-name-to-file 'eldav 'eldav-add-name-to-file)
 (put 'copy-file 'eldav 'eldav-copy-file)
@@ -1044,7 +1249,8 @@ Use lock TOKEN if specified."
 (put 'substitute-in-file-name 'eldav 'identity)
 (put 'unhandled-file-name-directory 'eldav
      'eldav-unhandled-file-name-directory)
-(put 'vc-registered 'eldav 'null)
+(put 'file-name-directory 'eldav 'eldav-file-name-directory)
+(put 'vc-registered 'eldav 'eldav-vc-registered)
 (put 'verify-visited-file-modtime 'eldav
      'eldav-verify-visited-file-modtime)
 (put 'expand-file-name 'eldav 'eldav-expand-file-name)
